/*
 GBxCart RW - Game Gear Flasher
 Version: 1.0
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 2/02/2023
 Last Modified: 2/02/2023
 License: CC-BY-NC-SA
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <signal.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

void sigintHandler(int sig_num) {
	set_mode('0');
	RS232_flushRX(cport_nr);
	gbx_cart_power_down();
	exit(0);
}

int main(int argc, char **argv) {
	signal(SIGINT, sigintHandler);
	
	printf("GBxCart RW - Game Gear 2MB Flasher v1.0 by insideGadgets\n");
	printf("#######################################################\n");
	
	// Check arguments
	if (argc >= 2) {
		read_config();
		
		// Open COM port
		if (com_test_port() == 0) {
			printf("Device not connected and couldn't be auto detected. Please make sure the COM port isn't open elsewhere. If you are using Linux/Mac, make sure you use \"sudo\".\n");
			read_one_letter();
			return 1;
		}
		printf("Connected on COM port: %i at ", cport_nr+1);
		
		// Break out of any existing functions on ATmega
		set_mode('0');
		RS232_flushRX(cport_nr);
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		// Get PCB version
		gbxcartPcbVersion = request_value(READ_PCB_VERSION);
		xmas_wake_up();
		
		// Increase to 1.7Mbit baud for v1.4 PCB
		#ifdef _WIN32
		if ((gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS) || gbxcartPcbVersion == MINI_1_1) {
			set_mode(USART_1_7M_SPEED);
			RS232_CloseComport(cport_nr);
			delay_ms(200);
			
			if (RS232_OpenComport(cport_nr, 1700000, "8N1") == 0) {
				printf("1.7M Baud\n");
			}
		}
		else {
			printf("1M Baud\n");
		}
		#else
			printf("1M Baud\n");
		#endif
		
		char pcbVersionString[20];
		if (gbxcartPcbVersion == PCB_1_0) {
			strncpy(pcbVersionString, "v1.0", 5);
		}
		else if (gbxcartPcbVersion == PCB_1_1) {
			strncpy(pcbVersionString, "v1.1/v1.2", 10);
		}
		else if (gbxcartPcbVersion == PCB_1_3) {
			strncpy(pcbVersionString, "v1.3", 5);
		}
		else if ((gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS)) {
			strncpy(pcbVersionString, "v1.4", 5);
		}
		else if (gbxcartPcbVersion == MINI) {
			strncpy(pcbVersionString, "Mini", 5);
		}
		else if (gbxcartPcbVersion == MINI_1_1) {
			strncpy(pcbVersionString, "Mini v1.1", 10);
		}
		else if (gbxcartPcbVersion == GBXMAS) {
			strncpy(pcbVersionString, "XMAS", 5);
		}
		
		// Get firmware version
		gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
		printf("PCB version: %s, Firmware version: %i\n", pcbVersionString, gbxcartFirmwareVersion);
		
		if (gbxcartPcbVersion == PCB_1_0) {
			printf("\nPCB v1.0 is not supported for this function.");
			read_one_letter();
			return 1;
		}
		if (gbxcartFirmwareVersion == 0) {
			printf("\nFirmware L1 may be installed. In order to use this applications you will need to downgrade to R30.\n");
			read_one_letter();
			return 1;
		}
		if (gbxcartFirmwareVersion <= 8) {
			printf("Firmware R9 or higher is required for this functionality.\n");
			read_one_letter();
			return 1;
		}
		
		#ifdef TIMETEST
			// Timing test
			printf("Timing started\n");
			clock_t t;
			t = clock();
		#endif
		
		// Separate the file name
		char filenameOnly[100];
		char filename[254];
		strncpy(filename, argv[1], 253);
		
#ifdef _WIN32
		char *token = strtok(filename, "\\");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "\\");
		}
#else
		char *token = strtok(filename, "/");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "/");
		}
#endif
		
		
		printf("\nROM File selected: %s\n\n", filenameOnly);
		
		uint32_t readBytes = 0;
		long fileSize = 0;
		
		// Grab file size
		FILE *romFile = fopen(argv[1], "rb");
		if (romFile != NULL) {
			fseek(romFile, 0, SEEK_END);
			fileSize = ftell(romFile);
			fseek(romFile, 0, SEEK_SET);
		}
		else {
			printf("\n%s \nFile not found\n", argv[1]);
			read_one_letter();
			return 1;
		}
	
		printf("insideGadgets 2 MByte Game Gear Flash Cart\n");
		
		if (fileSize > 0x200000) {
			fclose(romFile);
			printf("\n%s \nFile size is larger than the available Flash cart space of 2 MByte\n", argv[1]);
			read_one_letter();
			return 1;
		}
		
		printf("\nGoing to write %s\n", filenameOnly);
		
		// PCB v1.3 - Set 5V
		if (gbxcartPcbVersion == PCB_1_3 || (gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS) || gbxcartPcbVersion == GBXMAS) {
			set_mode(VOLTAGE_5V);
			delay_ms(500);
			gbx_cart_power_up();
		}
		
		currAddr = 0x0000;
		endAddr = 0xBFFF;
		
		// Calculate banks needed from ROM file size
		romBanks = ceil((double) fileSize / (double) 16384);
		if (romBanks <= 2) {
			romBanks = 3;
			endAddr = 0x7FFF;
		}
		
		// GG Init
		set_mode(GG_INIT);
		
		// Flash Setup
		set_mode(GB_CART_MODE); // Gameboy mode
		gb_flash_pin_setup(WE_AS_WR_PIN); // WR pin
		
		gb_flash_program_setup(GB_FLASH_PROGRAM_AAA);// Flash program byte method
		gb_check_change_flash_id(GB_FLASH_PROGRAM_AAA);
		
		// Chip erase
		printf("\nErasing Flash");
		xmas_chip_erase_animation();
		gb_flash_write_address_byte(0xAAA, 0xAA);
		gb_flash_write_address_byte(0x555, 0x55);
		gb_flash_write_address_byte(0xAAA, 0x80);
		gb_flash_write_address_byte(0xAAA, 0xAA);
		gb_flash_write_address_byte(0x555, 0x55);
		gb_flash_write_address_byte(0xAAA, 0x10);
		
		// Wait for first byte to be 0xFF
		wait_for_flash_chip_erase_ff(1);
		xmas_setup((romBanks * 16384) / 28);
		
		printf("\n\nWriting to ROM (Flash cart) from %s\n", filenameOnly);
		printf("[             25%%             50%%             75%%            100%%]\n[");
		
		// Write ROM
		currAddr = 0x0000;
		for (uint16_t bank = 2; bank < romBanks; bank++) {				
			if (bank > 2) { currAddr = 0x8000; }
			
			// Set start address
			set_number(currAddr, SET_START_ADDRESS);
			delay_ms(5);
			
			// Read data
			while (currAddr < endAddr) {
				if (currAddr == 0x8000) { // Switch banks here just before the next bank, not any time sooner
					set_bank(0xFFFF, bank);
					//printf("bank switched %i\n", bank);
					//read_one_letter();
				}
				
				com_write_bytes_from_file(GB_FLASH_WRITE_64BYTE, romFile, 64);
				com_wait_for_ack();
				currAddr += 64;
				readBytes += 64;
				
				// Print progress
				print_progress_percent(readBytes, (romBanks * 16384) / 64);
				led_progress_percent(readBytes, (romBanks * 16384) / 28);
				
				//printf("%i\n", readBytes);
			}
			//printf("%i\n", readBytes);
		}
		
		printf("]");
		fclose(romFile);
		
		gbx_set_done_led();
		gbx_cart_power_down();
		
		printf("\n\nCompleted.\n");
	}
	else {
		printf("\n*** Please drag and drop your rom file to this program***\n");
	}
	
	read_one_letter();
	
	
	return 0;
}
